/* Copyright 2004 Hiroyuki OYAMA. Japan. All rights reserved
 *
 */
/*
 * mod_auth_form: HTML Form based authentication
 *
 *
 * AuthForm directive
 *   Syntax:   AuthForm On|Off
 *   Default:  AuthForm Off
 *   Context:  directory, .htaccess
 *   Override: AuthConfig
 *
 *
 * AuthFormPasswordFile directive
 *   Syntax:   AuthFormPasswordFile <file-path>
 *   Context:  directory, .htaccess
 *   Override: AuthConfig
 *
 *
 * AuthFormLoginPage directive
 *   Syntax:   AuthFormLoginPage <file-path>
 *   Context:  directory, .htaccess
 *   Override: AuthConfig
 *
 *
 * AuthFormExpire directive
 *   Syntax:   AuthFormExpire <seconds>
 *   Default:  AuthFormExpire 28800
 *             #              8 hour
 *   Context:  directory, .htaccess
 *   Override: AuthConfig
 *
 *
 * AuthFormSecretKey directive
 *   Syntax:   AuthFormSecretKey <strings>
 *   Default:  AuthFormSecretKey "mod_auth_form.c default secret key"
 *   Context:  directory, .htaccess
 *   Override: AuthConfig
 * 
 */
#include "httpd.h"
#include "http_config.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"
#include "ap_config.h"
#include "ap_sha1.h"

#include <stdlib.h>

static const char DEFAULT_COOKIE_NAME[] = "mod_auth_form";
static const int  DEFAULT_EXPIRE        = 8 * 60 * 60;      /* 8 Hour */
static const char DEFAULT_SECRET[]      = "mod_auth_form.c default secret key";
static const char DEFAULT_LOGIN_FORM[] =
"<html>\n"
"  <head><title>Auth Required</title></head>\n"
"  <body>\n"
"    <h1>Auth Required</h1>\n"
"    <form method=\"POST\" action=\"apache_security_check\">\n"
"      <b>Login ID</b>: <input type=\"text\" name=\"apache_username\"><br>\n"
"      <b>Password</b>: <input type=\"password\" name=\"apache_password\"><br>\n"
"      <input type=\"submit\" value=\" sign on \">\n"
"    </form>\n"
"  </body>\n"
"</html>";
static const char FORM_ACTION[] = "apache_security_check";
static const char FORM_USERNAME[] = "apache_username";
static const char FORM_PASSWORD[] = "apache_password";


module auth_form_module;

typedef struct {
    int enable;
    const char *login_form;
    int expire;
    const char *secret;
    const char *cookie_name;
    char *pwfile;
} auth_form_conf;


static void *create_dir_conf(pool *p, char *dir)
{
    auth_form_conf *conf = ap_palloc(p, sizeof(auth_form_conf));

    conf->enable      = 0;
    conf->login_form  = NULL;
    conf->expire      = DEFAULT_EXPIRE;
    conf->secret      = DEFAULT_SECRET;
    conf->cookie_name = DEFAULT_COOKIE_NAME;
    conf->pwfile = NULL;

    return (void *)conf;
}

static const char *cmd_set_expire(cmd_parms *parms, void *mconf, char *expire_str)
{
    auth_form_conf *conf = mconf;

    conf->expire = atoi(expire_str);

    return NULL;
}

static const char *cmd_set_passwd_file(cmd_parms *parms, void *mconf, char *path)
{
    auth_form_conf *conf = mconf;

    conf->pwfile = path;

    return NULL;
}

static const command_rec cmds[] =
{
    { "AuthForm", ap_set_flag_slot,
      (void *)XtOffsetOf(auth_form_conf, enable), OR_AUTHCFG, TAKE1,
      "set HTML form authorization" },
    { "AuthFormLoginPage", ap_set_string_slot,
      (void *)XtOffsetOf(auth_form_conf, login_form), OR_AUTHCFG, TAKE1,
      "login form page's URI" },
    { "AuthFormExpire", cmd_set_expire,
      NULL, OR_AUTHCFG, TAKE1,
      "keep alive login cookie" },
    { "AuthFormSecretKey", ap_set_string_slot,
      (void *)XtOffsetOf(auth_form_conf, secret), OR_AUTHCFG, TAKE1,
      "set session secret key" },
    { "AuthFormPasswordFile",  cmd_set_passwd_file,
       NULL, OR_AUTHCFG, TAKE1,
      "The name of the file containing the usernames and password" },
    { NULL }
};


static table *parse_cookie(request_rec *r)
{
    const char *raw_cookie, *token, *key, *value;
    table *cookie;

    cookie = ap_make_table(r->pool, 1);
    raw_cookie = ap_table_get(r->headers_in, "Cookie");
    if (raw_cookie == NULL)
        return cookie;

    while (*raw_cookie && (token = ap_getword(r->pool, &raw_cookie, ' '))) {
         char *semi;
	 key   = ap_getword(r->pool, &token, '=');
         value = ap_getword(r->pool, &token, '=');
         if ((semi = strrchr(value, ';')) != NULL)
             *semi = '\0';
         ap_table_set(cookie, key, value);
    }
    return cookie;
}

static char *get_hmac_string(pool *p, const char *msg, const unsigned char *key)
{
    AP_SHA1_CTX ctx;
    unsigned char digest[SHA_DIGESTSIZE];
    unsigned char k_ipad[65], k_opad[65];
    unsigned char kt[SHA_DIGESTSIZE];
    unsigned char *k = (unsigned char *)key;
    char hex_digest[SHA_DIGESTSIZE * 2], *hd;
    const char hex[] = "0123456789abcdef";
    int key_len, i;

    key_len = strlen(key);
    /* if key is longer than 64 bytes reset it to key=SHA1(key) */
    if (key_len > 64) {
        k = kt;
        key_len = SHA_DIGESTSIZE; 
    }

    /* HMAC_MD5 transform */
    /* start out by string key in pads */
    memset((void *)k_ipad, 0, sizeof(k_ipad));
    memset((void *)k_opad, 0, sizeof(k_opad));
    memmove(k_ipad, k, key_len);
    memmove(k_opad, k, key_len);

    /* XOR key with ipad and opad values */
    for (i = 0; i < 64; i++) {
        k_ipad[i] ^= 0x36;
        k_opad[i] ^= 0x5c;
    }
    /* perform inner SHA-1 */
    ap_SHA1Init(&ctx);
    ap_SHA1Update_binary(&ctx, k_ipad, 64);
    ap_SHA1Update_binary(&ctx, (const unsigned char *)msg, strlen(msg));
    ap_SHA1Final(digest, &ctx);
    /* perform outer SHA-1 */
    ap_SHA1Init(&ctx);
    ap_SHA1Update_binary(&ctx, k_opad, 64);
    ap_SHA1Update_binary(&ctx, digest, sizeof(digest));
    ap_SHA1Final(digest, &ctx);

    hd = hex_digest;
    for (i = 0; i < sizeof(digest); i++) {
        *hd++ = hex[digest[i] >> 4];
        *hd++ = hex[digest[i] & 0xF];
    }
    *hd = '\0';

    return ap_pstrdup(p, hex_digest);
}

static int set_remote_user(request_rec *r, const char *user)
{
    const char *encoded, *auth_line;

    r->connection->user = ap_pstrdup(r->connection->pool, user);
    r->connection->ap_auth_type = "Basic";

    encoded = ap_pbase64encode(r->pool, ap_pstrcat(r->pool, user, ":", NULL));
    auth_line = ap_pstrcat(r->pool, "Basic ", encoded, NULL);
    ap_table_set(r->headers_in, "Authorization", auth_line);

    return 1;
}

static int is_valid_cookie(request_rec *r, table *cookie)
{
    auth_form_conf *conf;
    const char *cookie_value;
    const char *expstr, *user, *hmac, *myhmac;
    time_t expire;

    conf = ap_get_module_config(r->per_dir_config, &auth_form_module);
    cookie_value = ap_table_get(cookie, conf->cookie_name);
    if (cookie_value == NULL) {
        return 0;
    }
    if (strlen(cookie_value) <= SHA_DIGESTSIZE * 2) {
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
                      "auth cookie is shorten: %s",
                      ap_escape_logitem(r->pool, cookie_value));
        return 0;
    }
    hmac   = ap_getword(r->pool, &cookie_value, ',');
    user   = ap_getword(r->pool, &cookie_value, ',');
    expstr = ap_getword(r->pool, &cookie_value, ',');
    if (hmac == NULL || user == NULL || expstr == NULL
        || *hmac == '\0' || *user == '\0' || *expstr == '\0') {
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
                      "auth cookie is invalid format: %s",
                      ap_escape_logitem(r->pool, cookie_value));
        return 0;
    }

    /* check HMAC */
    myhmac = get_hmac_string(r->pool,
                             ap_psprintf(r->pool, ",%s,%s", user, expstr),
                             conf->secret);
    if (strcmp(hmac, myhmac) != 0) {
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
                      "auth cookie has invalid hmac: %s",
                      ap_escape_logitem(r->pool, cookie_value));
        return 0;
    }

    /* check expire */
    expire = atoi(expstr);    
    if (expire == NULL || r->request_time >= expire) {
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_WARNING, r,
                      "auth cookie is expired: %s",
                      ap_escape_logitem(r->pool, cookie_value));
        return 0;
    }

    /* set username */
    set_remote_user(r, user);

    return 1;
}



static int set_login_cookie(request_rec *r)
{
    auth_form_conf *conf;
    const char *cookie, *src, *hmac;
    time_t expire;

    conf = ap_get_module_config(r->per_dir_config, &auth_form_module);
    expire = time(NULL) + conf->expire; 
    src = ap_psprintf(r->pool, ",%s,%u",
                      r->connection->user, (unsigned int)expire);
    hmac = get_hmac_string(r->pool, src, conf->secret); 
    
    cookie = ap_pstrcat(r->pool, conf->cookie_name, "=", hmac, src, ";", NULL);
    ap_table_set(r->headers_out, "Set-Cookie", cookie);
    ap_table_set(r->err_headers_out, "Set-Cookie", cookie);

    ap_table_set(r->headers_out, "Location", r->unparsed_uri);
    return REDIRECT;
}


static const char *load_file(request_rec *r, const char *path)
{
    struct stat finfo;
    int fd;
    char *body;
    ssize_t rc;

    if (stat(path, &finfo) == -1) {
        return DEFAULT_LOGIN_FORM;
    }
    fd = ap_popenf(r->pool, path, O_RDONLY, 0);
    if (fd < 0) {
        return DEFAULT_LOGIN_FORM;
    }

    body = ap_palloc(r->pool, finfo.st_size);
    rc = read(fd, body, finfo.st_size);
    ap_pclosef(r->pool, fd);

    return body;
}

static const char *parse_by_body(request_rec *r)
{
    char *body;
    int rc;
    long length;

    if (ap_setup_client_block(r, REQUEST_CHUNKED_ERROR))
        return "";
    if (!ap_should_client_block(r))
        return "";

    length = r->remaining;
    body = ap_pcalloc(r->pool, length + 1);
    rc = ap_get_client_block(r, body, length);

    return (const char *)body;
}

static void plustospace(char *str)
{
    register int x;
    for(x=0;str[x];x++) if(str[x] == '+') str[x] = ' ';
}


static table *parse_request_param(request_rec *r)
{
    table *param = ap_make_table(r->pool, 1);
    const char *args, *chunk;

    args = ap_pstrcat(r->pool, parse_by_body(r), "&", args, NULL);

    while (*args && (chunk = ap_getword(r->pool, &args, '&'))) {
        char *key, *value;

        key = ap_getword(r->pool, &chunk, '=');
        value = ap_getword(r->pool, &chunk, '=');
        plustospace(key);
        plustospace(value);
        ap_unescape_url(key);
        ap_unescape_url(value);
        ap_table_set(param, key, value);
    }

    return param;
}

static int show_login_form(request_rec *r)
{
    auth_form_conf *conf;
    const char *login_form;
    char *find;

    conf = ap_get_module_config(r->per_dir_config, &auth_form_module);

    r->no_cache = 1;
    if (conf->login_form == NULL) {
        login_form = DEFAULT_LOGIN_FORM;
    }
    else {
        login_form = load_file(r, conf->login_form);
    }

    r->content_type = "text/html";
    r->no_cache = 1;
    ap_send_http_header(r);

    find = strstr(login_form, FORM_ACTION);
    if (find == NULL) {
        ap_rputs(login_form, r);
        return DONE;
    }

    ap_rwrite(login_form, find - login_form, r);
    ap_rputs(ap_escape_html(r->pool, r->unparsed_uri), r);
    ap_rputs(find + strlen(FORM_ACTION), r);

    return DONE;
}

static char *get_password(request_rec *r, const char *user, const char *path)
{
    configfile_t *f;
    char l[MAX_STRING_LEN];
    const char *rpw, *w;

    if (!(f = ap_pcfg_openfile(r->pool, path))) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
                      "Could not open password file: %s", path);
        return NULL;
    }
    while (!(ap_cfg_getline(l, MAX_STRING_LEN, f))) {
        if ((l[0] == '#') || (!l[0]))
            continue;
        rpw = l;
        w = ap_getword(r->pool, &rpw, ':');

        if (!strcmp(user, w)) {
            ap_cfg_closefile(f);
            return ap_getword(r->pool, &rpw, ':');
        }
    }
    ap_cfg_closefile(f);
    return NULL;
}

static int check_auth(request_rec *r)
{
    auth_form_conf *conf;
    table *cookie, *query;
    const char *user, *passwd;
    char *real_pw, *rc;

    if (!ap_is_initial_req(r)) {
        return DECLINED;
    }

    conf = ap_get_module_config(r->per_dir_config, &auth_form_module);
    if (!conf->enable)
        return DECLINED;

    cookie = parse_cookie(r);
    if (is_valid_cookie(r, cookie)) {
        return OK;
    }

    if (r->method_number != M_POST) {
        return show_login_form(r);
    }

    query = parse_request_param(r);
    user = ap_table_get(query, FORM_USERNAME);
    passwd = ap_table_get(query, FORM_PASSWORD);
    if (user == NULL)
        user = "";
    if (passwd == NULL)
        passwd = "";

    real_pw = get_password(r, user, conf->pwfile);
    if (real_pw == NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
                      "user %s not found: %s", user, r->uri);
        return show_login_form(r);
    }

    rc = ap_validate_password(passwd, real_pw);
    if (rc != NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
                      "user \"%s\": authentication failure for \"%s\": %s",
                      user, r->uri, rc);
        return show_login_form(r);
    }

    set_remote_user(r, user);
    return set_login_cookie(r);
}

/* Dispatch list for API hooks */
module MODULE_VAR_EXPORT auth_form_module = {
    STANDARD_MODULE_STUFF, 
    NULL,                  /* module initializer                  */
    create_dir_conf,       /* create per-dir    config structures */
    NULL,                  /* merge  per-dir    config structures */
    NULL,                  /* create per-server config structures */
    NULL,                  /* merge  per-server config structures */
    cmds,                  /* table of config file commands       */
    NULL,                  /* [#8] MIME-typed-dispatched handlers */
    NULL,                  /* [#1] URI to filename translation    */
    NULL,                  /* [#4] validate user id from request  */
    NULL,                  /* [#5] check if the user is ok _here_ */
    check_auth,            /* [#3] check access by host address   */
    NULL,                  /* [#6] determine MIME type            */
    NULL,                  /* [#7] pre-run fixups                 */
    NULL,                  /* [#9] log a transaction              */
    NULL,                  /* [#2] header parser                  */
    NULL,                  /* child_init                          */
    NULL,                  /* child_exit                          */
    NULL                   /* [#0] post read-request              */
#ifdef EAPI
   ,NULL,
    NULL,
    NULL,
    NULL
#endif
};

